<?php

declare(strict_types=1);

namespace Drupal\simple_oauth_token_exchange\EventSubscriber;

use Drupal\Core\Url;
use Drupal\simple_oauth\Entities\ScopeEntityInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Response subscriber to add RFC8693 parameters to token exchange responses.
 */
class TokenExchangeResponseSubscriber implements EventSubscriberInterface {

  public function __construct(
    protected readonly RequestStack $requestStack,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    return [
      KernelEvents::RESPONSE => ['onResponse', -100],
    ];
  }

  /**
   * Handles the response event.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The response event.
   */
  public function onResponse(ResponseEvent $event): void {
    $request = $event->getRequest();
    $response = $event->getResponse();

    // Only process token endpoint requests with grant_type=token_exchange.
    if (!$this->isTokenExchangeRequest($request)) {
      return;
    }

    $content = json_decode($response->getContent(), TRUE);
    if (!is_array($content)) {
      return;
    }

    // Only modify successful responses.
    if ($response->getStatusCode() !== 200) {
      return;
    }

    // Add the issued token type.
    $content['issued_token_type'] = 'urn:ietf:params:oauth:token-type:access_token';

    // Add the issued token scopes.
    // The scopes are stored in the current request in the
    // grant type for use here.
    if (!isset($content['scope'])) {
      /** @var \Drupal\simple_oauth\Entities\ScopeEntityInterface[] $finalScopes */
      $finalScopes = $this->requestStack->getCurrentRequest()->attributes->get('oauth2_token_exchange_final_scopes');

      $content['scope'] = implode(' ', array_map(function (ScopeEntityInterface $scope) {
        return $scope->getIdentifier();
      }, $finalScopes));
    }

    $response->setContent(json_encode($content));
  }

  /**
   * Checks if the request is a token exchange request.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   *
   * @return bool
   *   TRUE if this is a token exchange request.
   */
  protected function isTokenExchangeRequest($request): bool {
    $path = Url::fromRoute('oauth2_token.token')->setAbsolute(FALSE)->toString();

    return $request->getPathInfo() === $path
      && $request->request->get('grant_type') === 'urn:ietf:params:oauth:grant-type:token-exchange';
  }

}
