<?php

declare(strict_types=1);

namespace Drupal\simple_oauth_token_exchange\Grant;

use Drupal\simple_oauth_token_exchange\Service\SubjectTokenManagerFactory;
use Drupal\simple_oauth_token_exchange\TokenTypeIdentifiers;
use League\OAuth2\Server\Grant\AbstractGrant;
use League\OAuth2\Server\Grant\GrantTypeInterface;
use League\OAuth2\Server\ResponseTypes\ResponseTypeInterface;
use League\OAuth2\Server\Repositories\RefreshTokenRepositoryInterface;
use League\OAuth2\Server\Exception\OAuthServerException;
use Psr\Http\Message\ServerRequestInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Implements RFC 8693 OAuth 2.0 Token Exchange.
 *
 * @see https://datatracker.ietf.org/doc/html/rfc8693
 */
class TokenExchangeGrant extends AbstractGrant implements GrantTypeInterface {

  /**
   * {@inheritdoc}
   */
  public function __construct(
    RefreshTokenRepositoryInterface $refreshTokenRepository,
    protected readonly SubjectTokenManagerFactory $subjectTokenManagerFactory,
    protected readonly RequestStack $requestStack,
  ) {
    $this->setRefreshTokenRepository($refreshTokenRepository);
    $this->refreshTokenTTL = new \DateInterval('P1M');
  }

  /**
   * {@inheritdoc}
   */
  public function getIdentifier(): string {
    return 'urn:ietf:params:oauth:grant-type:token-exchange';
  }

  /**
   * {@inheritdoc}
   */
  public function respondToAccessTokenRequest(
    ServerRequestInterface $request,
    ResponseTypeInterface $responseType,
    \DateInterval $accessTokenTTL,
  ): ResponseTypeInterface {
    // Validate request.
    $client = $this->validateClient($request);
    $scopes = $this->validateScopes($this->getRequestParameter('scope', $request));

    // Validate the subject token (required)
    $subjectToken = $this->getRequestParameter('subject_token', $request);
    $this->validateSubjectToken($subjectToken);

    $subjectTokenType = $this->getRequestParameter('subject_token_type', $request);
    $this->validateSubjectTokenType($subjectTokenType);

    // Check for unsupported parameters.
    $this->checkForUnsupportedParameters($request);

    $subjectTokenManager = $this->subjectTokenManagerFactory->create($subjectToken);

    // Validate the subject token (validity, client and scopes).
    $subjectTokenManager->validate($client->getIdentifier(), $scopes);

    // Make sure the subject token is valid and has the requested scopes.
    // Additionally makes sure that each scope is allowed for the
    // token exchange.
    $oauthToken = $subjectTokenManager->getTokenEntity();

    $finalScopes = $subjectTokenManager->getFinalScopes($scopes);

    // Generate new access token.
    $accessToken = $this->issueAccessToken(
      $accessTokenTTL,
      $client,
      $oauthToken->get('auth_user_id')->getString(),
      $finalScopes,
    );

    // Store the final scopes for downstream use.
    $this->requestStack->getCurrentRequest()->attributes->set('oauth2_token_exchange_final_scopes', $finalScopes);

    // Send the response.
    $responseType->setAccessToken($accessToken);

    return $responseType;
  }

  /**
   * Validates the subject token.
   *
   * @param string|null $subjectToken
   *   The subject token.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   If the subject token is invalid.
   */
  protected function validateSubjectToken(?string $subjectToken): void {
    if (!$subjectToken) {
      throw OAuthServerException::invalidRequest('subject_token', 'The subject token is required.');
    }
  }

  /**
   * Validates the subject token type.
   *
   * This module only supports access tokens as subject tokens.
   *
   * @param string $subjectTokenType
   *   The subject token type.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   If the subject token type is invalid.
   */
  protected function validateSubjectTokenType(string $subjectTokenType): void {
    if (empty($subjectTokenType)) {
      throw OAuthServerException::invalidRequest('subject_token_type', 'The subject token type is required.');
    }

    if ($subjectTokenType !== TokenTypeIdentifiers::ACCESS_TOKEN) {
      throw OAuthServerException::invalidRequest('subject_token_type', 'The subject token type must be an access token.');
    }
  }

  /**
   * Checks for unsupported parameters.
   *
   * @param \Psr\Http\Message\ServerRequestInterface $request
   *   The request.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   If an unsupported parameter is found.
   */
  protected function checkForUnsupportedParameters(ServerRequestInterface $request): void {
    // Validate requested token type.
    // This parameter is not used for anything, because only access tokens are
    // supported as subject tokens, but the setting and validation are required
    // by the specification.
    $requestedTokenType = $this->getRequestParameter('requested_token_type', $request);
    if ($requestedTokenType && $requestedTokenType !== TokenTypeIdentifiers::ACCESS_TOKEN) {
      throw OAuthServerException::invalidRequest('requested_token_type', 'The requested token type must be an access token.');
    }
    // The actor_token and actor_token_type parameters are not supported.
    $actorToken = $this->getRequestParameter('actor_token', $request);
    $actorTokenType = $this->getRequestParameter('actor_token_type', $request);
    if ($actorToken || $actorTokenType) {
      throw OAuthServerException::invalidRequest('actor_token', 'The actor_token and actor_token_type parameters are not supported.');
    }

    // The resource parameter is not supported.
    $resource = $this->getRequestParameter('resource', $request);
    if ($resource) {
      throw OAuthServerException::invalidRequest('resource', 'The resource parameter is not supported. Please set the desired scopes instead.');
    }
  }

}
