<?php

declare(strict_types=1);

namespace Drupal\simple_oauth_token_exchange\Service;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\simple_oauth\Entities\ScopeEntity;
use Drupal\simple_oauth\Entity\Oauth2TokenInterface;
use GuzzleHttp\Psr7\ServerRequest;
use League\OAuth2\Server\Entities\ScopeEntityInterface;
use League\OAuth2\Server\Exception\OAuthServerException;
use League\OAuth2\Server\Repositories\ScopeRepositoryInterface;
use League\OAuth2\Server\ResourceServer;

/**
 * Service for managing subject tokens in token exchange.
 */
class SubjectTokenManager {

  /**
   * The token entity.
   */
  protected ?Oauth2TokenInterface $token = NULL;

  /**
   * The final scopes.
   */
  protected array $finalScopes = [];

  /**
   * Constructs a new SubjectTokenManager.
   */
  public function __construct(
    protected readonly string $tokenValue,
    protected readonly ResourceServer $resourceServer,
    protected readonly ScopeRepositoryInterface $scopeRepository,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
  ) {}

  /**
   * Validates a subject token.
   *
   * @param string $requestedClientId
   *   The client ID that requested the token exchange.
   * @param array<\Drupal\simple_oauth\Entities\ScopeEntityInterface> $requestedScopes
   *   Optional array of requested scopes. These scopes
   *   must be present on the subject token.
   *
   * @return \Drupal\simple_oauth\Entity\Oauth2TokenInterface
   *   The validated token entity.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   When the token is invalid, expired, or revoked.
   */
  public function validate(string $requestedClientId, array $requestedScopes = []): ?Oauth2TokenInterface {
    try {
      $tokenEntity = $this->getTokenEntity();

      // Check if the token has been revoked.
      if ($tokenEntity->isRevoked()) {
        throw OAuthServerException::invalidRequest('subject_token', 'Token has been revoked.');
      }

      // Validate client.
      $this->validateClient($tokenEntity, $requestedClientId);

      // Validate scopes.
      $this->validateScopes($tokenEntity, $requestedScopes);

      return $tokenEntity;
    }
    catch (OAuthServerException $e) {
      throw $e;
    }
    catch (\Exception $e) {
      throw OAuthServerException::invalidRequest('subject_token', 'Invalid token format.');
    }
  }

  /**
   * Validates the scopes for the token exchange.
   *
   * @param \Drupal\simple_oauth\Entity\Oauth2TokenInterface $tokenEntity
   *   The token entity.
   * @param array<\Drupal\simple_oauth\Entities\ScopeEntityInterface> $requestedScopes
   *   The requested scopes.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   When the scopes are invalid.
   */
  protected function validateScopes(Oauth2TokenInterface $tokenEntity, array $requestedScopes): void {
    $finalScopes = $this->getFinalScopes($requestedScopes);

    // If requested scopes are provided, make sure
    // all required scopes are present on the token.
    if (!empty($requestedScopes)) {
      $subjectTokenScopes = $this->getTokenScopes($tokenEntity);
      $subjectTokenScopeIds = array_map(fn (ScopeEntityInterface $scope) => $scope->getIdentifier(), $subjectTokenScopes);

      foreach ($finalScopes as $scope) {
        if (!in_array($scope->getIdentifier(), $subjectTokenScopeIds)) {
          throw OAuthServerException::invalidRequest(
            'scope',
            'The following requested scope must be present on the subject token but is missing: ' . $scope->getIdentifier()
          );
        }
      }
    }

    // Now make sure that each scope is allowed
    // for the token exchange grant type.
    foreach ($finalScopes as $scope) {
      if ($scope instanceof ScopeEntity && !$scope->getScopeObject()->isGrantTypeEnabled('token_exchange')) {
        throw OAuthServerException::invalidScope($scope->getIdentifier());
      }
    }
  }

  /**
   * Makes sure the token belongs to the requested client.
   *
   * @param \Drupal\simple_oauth\Entity\Oauth2TokenInterface $tokenEntity
   *   The token entity.
   * @param string $requestedClientId
   *   The client ID that requested the token exchange.
   *
   * @throws \League\OAuth2\Server\Exception\OAuthServerException
   *   When the client ID is invalid.
   */
  protected function validateClient(Oauth2TokenInterface $tokenEntity, string $requestedClientId): void {
    $clientStorage = $this->entityTypeManager->getStorage('consumer');
    /** @var \Drupal\consumers\Entity\Consumer $clientEntity */
    $clientEntity = $clientStorage->load($tokenEntity->get('client')->getString());

    if (!$clientEntity || $clientEntity->getClientId() !== $requestedClientId) {
      throw OAuthServerException::invalidRequest('subject_token', 'Token does not belong to the requested client.');
    }
  }

  /**
   * Get the scopes entities for a token.
   *
   * @return array<\Drupal\simple_oauth\Entities\ScopeEntityInterface>
   *   The scope entities.
   */
  protected function getTokenScopes(Oauth2TokenInterface $tokenEntity): array {
    $scopeEntities = [];

    foreach ($tokenEntity->get('scopes')->getValue() as $scope) {
      $scopeEntities[] = $this->scopeRepository->getScopeEntityByIdentifier($scope['scope_id']);
    }

    return $scopeEntities;
  }

  /**
   * Get the final scopes for the token exchange.
   *
   * The final scopes are the requested scopes, or the
   * subject token's scopes if no requested scopes are provided.
   *
   * @param array<\Drupal\simple_oauth\Entities\ScopeEntityInterface> $requestedScopes
   *   The requested scopes.
   *
   * @return array<\Drupal\simple_oauth\Entities\ScopeEntityInterface>
   *   The final scopes.
   */
  public function getFinalScopes(array $requestedScopes): array {
    if (!$this->finalScopes) {
      $tokenEntity = $this->getTokenEntity();

      // Get the scopes for the token.
      $subjectTokenScopes = $this->getTokenScopes($tokenEntity);

      // If no required scopes are provided, use the token's scopes.
      $this->finalScopes = empty($requestedScopes) ? $subjectTokenScopes : $requestedScopes;
    }

    return $this->finalScopes;
  }

  /**
   * Get the token entity.
   *
   * Loads the token entity if it hasn't been loaded yet.
   *
   * @return \Drupal\simple_oauth\Entity\Oauth2TokenInterface
   *   The token entity.
   */
  public function getTokenEntity() {
    if (!$this->token) {
      // Create a server request with the token in the Authorization header.
      $serverRequest = (new ServerRequest('GET', '/', [
        'Authorization' => 'Bearer ' . $this->tokenValue,
      ]));

      // Validate the JWT token.
      $validatedRequest = $this->resourceServer->validateAuthenticatedRequest($serverRequest);

      // Get the token ID from the validated request.
      $tokenId = $validatedRequest->getAttribute('oauth_access_token_id');

      $tokenStorage = $this->entityTypeManager->getStorage('oauth2_token');
      $tokens = $tokenStorage->loadByProperties(['value' => $tokenId]);
      if (!$tokens) {
        throw OAuthServerException::invalidRequest('subject_token', 'Token not found.');
      }

      $this->token = reset($tokens);
    }

    return $this->token;
  }

}
