<?php

declare(strict_types=1);

namespace Drupal\Tests\simple_oauth_token_exchange\Functional;

use Drupal\consumers\Entity\Consumer;
use Drupal\Core\Url;
use Drupal\Tests\simple_oauth\Functional\RequestHelperTrait;

/**
 * Tests the token exchange functionality between different clients.
 *
 * @group simple_oauth_token_exchange
 */
class CrossClientTokenExchangeTest extends TokenExchangeFunctionalTestBase {

  use RequestHelperTrait;

  /**
   * The token URL.
   */
  protected Url $tokenUrl;

  /**
   * The password-only client.
   */
  protected Consumer $passwordClient;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->tokenUrl = Url::fromRoute('oauth2_token.token');

    // Create a client that supports both password and token exchange.
    $this->client = $this->createClient(['password', 'token_exchange']);

    // Create another client that only supports password grant.
    $this->passwordClient = $this->createClient(['password']);

    // Create scopes.
    $this->createScope('test_password', ['password']);
    $this->createScope('test_exchange', ['password', 'token_exchange']);
  }

  /**
   * Tests token exchange between different clients.
   */
  public function testCrossClientTokenExchange(): void {
    // First, get an access token from the password-only client.
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->passwordClient->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_password',
    ]);

    $this->assertEquals(200, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];
    $this->assertNotEmpty($accessToken);

    // Now try to exchange this token using the client that
    // supports token exchange.
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_exchange',
    ]);

    // The exchange should fail because the token is from a different client.
    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('Token does not belong to the requested client.', $body['hint']);
  }

  /**
   * Tests token exchange with a client that doesn't support token exchange.
   */
  public function testTokenExchangeWithUnsupportedClient(): void {
    // First, get an access token from the token exchange capable client.
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exchange',
    ]);

    $this->assertEquals(200, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];
    $this->assertNotEmpty($accessToken);

    // Try to use token exchange with the password-only client.
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->passwordClient->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_password',
    ]);

    // The request should fail because the client doesn't
    // support token exchange.
    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('unsupported_grant_type', $body['error']);
  }

}
