<?php

declare(strict_types=1);

namespace Drupal\Tests\simple_oauth_token_exchange\Functional;

use Drupal\Core\Url;
use Drupal\Tests\simple_oauth\Functional\RequestHelperTrait;

/**
 * Tests the token exchange functionality through HTTP requests.
 *
 * @group simple_oauth_token_exchange
 */
class TokenExchangeTest extends TokenExchangeFunctionalTestBase {

  use RequestHelperTrait;

  /**
   * The token URL.
   */
  protected Url $tokenUrl;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->tokenUrl = Url::fromRoute('oauth2_token.token');
    $this->client = $this->createClient(['password', 'token_exchange']);

    $this->createScope('test_exclusive', ['password']);
    $this->createScope('test_exchange', ['password', 'token_exchange']);
  }

  /**
   * Tests the token exchange endpoint.
   */
  public function testTokenExchange(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exclusive test_exchange',
    ]);

    $this->assertEquals(200, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);

    $accessToken = $body['access_token'];
    $this->assertNotEmpty($accessToken);

    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_exchange',
    ]);

    $this->assertEquals(200, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertNotEmpty($body['access_token']);
    $this->assertEquals($body['scope'], 'test_exchange');
  }

  /**
   * Tests token exchange without scope parameter.
   */
  public function testTokenExchangeWithoutScope(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exclusive test_exchange',
    ]);

    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];

    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
    ]);

    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('invalid_scope', $body['error']);
  }

  /**
   * Tests token exchange with exclusive scope.
   */
  public function testTokenExchangeWithExclusiveScope(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exclusive test_exchange',
    ]);

    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];

    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_exclusive',
    ]);

    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('invalid_scope', $body['error']);
  }

  /**
   * Tests token exchange with non-existent scope.
   */
  public function testTokenExchangeWithNonExistentScope(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exclusive test_exchange',
    ]);

    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];

    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'non_existent_scope',
    ]);

    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('invalid_scope', $body['error']);
  }

  /**
   * Tests token exchange with wrong subject token.
   */
  public function testTokenExchangeWithWrongSubjectToken(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => 'wrong_token',
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_exchange',
    ]);

    $this->assertEquals(401, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('access_denied', $body['error']);
  }

  /**
   * Tests token exchange with missing subject token.
   */
  public function testTokenExchangeWithMissingSubjectToken(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token_type' => 'urn:ietf:params:oauth:token-type:access_token',
      'scope' => 'test_exchange',
    ]);

    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('invalid_request', $body['error']);
    $this->assertStringContainsString('subject token is required.', $body['hint']);
  }

  /**
   * Tests token exchange with wrong subject token type.
   */
  public function testTokenExchangeWithWrongSubjectTokenType(): void {
    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'password',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'username' => $this->user->getAccountName(),
      // @phpstan-ignore property.notFound
      'password' => $this->user->pass_raw,
      'scope' => 'test_exclusive test_exchange',
    ]);

    $body = json_decode((string) $response->getBody(), TRUE);
    $accessToken = $body['access_token'];

    $response = $this->post($this->tokenUrl, [
      'grant_type' => 'urn:ietf:params:oauth:grant-type:token-exchange',
      'client_id' => $this->client->getClientId(),
      'client_secret' => $this->clientSecret,
      'subject_token' => $accessToken,
      'subject_token_type' => 'wrong_token_type',
      'scope' => 'test_exchange',
    ]);

    $this->assertEquals(400, $response->getStatusCode());
    $body = json_decode((string) $response->getBody(), TRUE);
    $this->assertEquals('invalid_request', $body['error']);
    $this->assertStringContainsString('subject token type must be an access token', $body['hint']);
  }

}
